import os

import sqlalchemy as sa
import pytest

from app import app as flask_app, DB
from models.agreement import Agreement
from pytest_postgresql.factories import init_postgresql_database, drop_postgresql_database

from models.user import User

DB_OPTS = {"user": "postgres", "host": "localhost", "port": 5432, "database": "test_2", "password": "postgres"}
PG_USER = "postgres"
PG_PASSWORD = "postgres"
PG_PORT = 5432
SQLALCHEMY_DATABASE_URI = "postgresql://{username}:{password}" \
                          "@localhost:{port}/flask_testing".format(username=PG_USER,
                                                                   password=PG_PASSWORD,
                                                                   port=PG_PORT)
"""
https://github.com/jeancochrane/pytest-flask-sqlalchemy/blob/master/tests/_conftest.py
"""

# Retrieve a database connection string from the shell environment
try:
    DB_CONN = os.environ['TEST_DATABASE_URL']
except KeyError:
    raise KeyError('TEST_DATABASE_URL not found. You must export a database ' +
                   'connection string to the environmental variable ' +
                   'TEST_DATABASE_URL in order to run tests.')
else:
    DB_OPTS = sa.engine.url.make_url(DB_CONN).translate_connect_args()

pytest_plugins = ['pytest-flask-sqlalchemy']


@pytest.fixture(scope='session')
def database(request):
    '''
    Create a Postgres database for the tests, and drop it when the tests are done.
    '''
    pg_host = DB_OPTS.get("host")
    pg_port = DB_OPTS.get("port")
    pg_user = DB_OPTS.get("username")
    pg_db = DB_OPTS["database"]

    init_postgresql_database(pg_user, pg_host, pg_port, pg_db, password="postgres")

    @request.addfinalizer
    def drop_database():
        drop_postgresql_database(pg_user, pg_host, pg_port, pg_db, 9.6, password="postgres")


SQLALCHEMY_DATABASE = None


@pytest.fixture(scope='session')
def app(database):
    '''
    Create a Flask app context for the tests.
    '''
    # _app.config.from_object('config.TestConfig')
    # flask_app.config['SQLALCHEMY_DATABASE_URI'] = DB_CONN

    yield flask_app


@pytest.fixture(scope='session')
def _db(app):
    """
    Provide the transactional fixtures with access to the database via a Flask-SQLAlchemy
    database connection.
    """
    DB.create_all()
    return DB


@pytest.fixture(scope="function")
def agreement_object(db_session):
    user = User(email='testing@mail.com')
    DB.session.add(user)

    expected = {'agreement_status': False, 'user_id': user.id}
    agreement = Agreement(**expected)
    DB.session.add(agreement)
    DB.session.commit()

    yield {"agreement": agreement.id, "user": user.id}



import os
from app import DB, DB_FILE_NAME
from models.agreement import AgreementDummy

# Delete database file if it exists currently
if os.path.exists(DB_FILE_NAME):
    os.remove(DB_FILE_NAME)

# Create the database
DB.create_all()

# create data in database
AgreementDummy.generate_dummy_agreements()


print("Data Import Successful.")



import logging

from connexion import decorators

logger = logging.getLogger("connexion.decorators.validation")


class RequestBodyValidator(decorators.validation.RequestBodyValidator):
    """
    This class overrides the default connexion RequestBodyValidator
    so that it returns the complete string representation of the
    error, rather than just returning the error message.

    For more information:
        - https://github.com/zalando/connexion/issues/558
        - https://connexion.readthedocs.io/en/latest/request.html
    """

    def validate_schema(self, data, url):
        return super(RequestBodyValidator, self).validate_schema(data, url)



from app import DB, MA


class UserDummy:
    @staticmethod
    def get_dummy_data():
        users_instances = [User(email=email) for email in ["user@one.com", "user@two.com", "user@three.com"]]
        [DB.session.add(user) for user in users_instances]
        DB.session.commit()
        return [user.id for user in users_instances]


class User(DB.Model):
    __tablename__ = "user"  # noqa

    id = DB.Column(DB.Integer, primary_key=True)
    email = DB.Column(DB.String, unique=True, nullable=False)
    # agreement = DB.relationship("Agreement",
    #                             backref=DB.backref('user', lazy='joined'),
    #                             uselist=False)  # use_list=False to make one-to-one


class UserSchema(MA.SQLAlchemyAutoSchema):
    class Meta:
        model = User
        sqla_session = DB.session  # noqa
        # include_relationships = False
        # include_fk = True



BASE_API_URL = "/api"
AGREEMENT_RESOURCE_PATH = BASE_API_URL + "/agreement"


def test_get_agreement_endpoint(client, agreement_object, db_session):
    response = client.get(AGREEMENT_RESOURCE_PATH, content_type='application/json')
    assert response.status_code == 200
    assert all(response.json[0].get(key) == agreement_object[key] for key in agreement_object.keys())
    assert len(response.json) == 1


def test_add_agreement_endpoint(client, agreement_object, db_session):
    user_id = agreement_object.get("user")
    expected = {'agreement_status': False, 'user_id': user_id}
    response = client.post(AGREEMENT_RESOURCE_PATH, json={
        'user_id': user_id, "agreement_status": False
    })
    assert response.status_code == 201
    assert all(response.json.get(key) == value for key, value in expected.items())



from werkzeug.exceptions import abort

from app import DB
from models.agreement import Agreement, AgreementSchema
from models.user import User
from models.utils import get_timestamp


# Create a handler for our read (GET) agreements
def read_agreements():
    """
    This function responds to a request for /api/agreement
    with the complete lists of agreements.

    :return:        sorted list of agreements
    """
    # Create the list of people from our data
    agreements = Agreement.query.order_by(Agreement.agreement_status).all()

    # Serialize the data for the response
    agreement_schema = AgreementSchema(many=True)
    return agreement_schema.dump(agreements), 200


def agreement_form_dict(body):
    agreement_schema = AgreementSchema()
    de_serialized_agreement = agreement_schema.load(body, session=DB.session)
    return Agreement(**de_serialized_agreement)


def add_agreement(body):
    """
    This function creates a new agreement in the agreement structure
    based on the passed-in agreement data

    :param body:  agreement to create in agreement structure
    :return:        201 on success, 406 on agreement exists
    """
    agreement_already_exists = Agreement.query.filter(Agreement.user_id == body.get("user_id")).one_or_none()
    user = User.query.filter_by(id=body.get("user_id")).one_or_none()
    if not user:
        abort(400, "Specified user does't exits"),

    if agreement_already_exists:
        abort(409, f"Agreement against {body.get('user_id')} exists already")

    agreement = agreement_form_dict(body)
    DB.session.add(agreement)
    DB.session.commit()

    serialized = AgreementSchema().dump(agreement)

    return serialized, 201  # created


def get_agreement(agreement_id):
    agreement = Agreement.query.filter(Agreement.id == agreement_id).one_or_none()
    if not agreement:
        abort(404, "Invalid Agreement id.")

    agreement_schema = AgreementSchema()
    serialized_agreement = agreement_schema.dump(agreement)
    return serialized_agreement


def accept_agreement(uuid):
    """
    This function only marks an existing agreement as accepted.

    :param uuid:  uuid of target agreement.
    :return:        201 on success, 404 if agreement does't exists
    """

    agreement = Agreement.query.filter_by(uuid=uuid).first()

    if not agreement:
        abort(404, f"Invalid agreement uuid.")

    agreement.agreement_status = True
    agreement.agreed_on = get_timestamp()
    # DB.session.add(agreement) # not required in case of update.
    DB.session.commit()

    return AgreementSchema().dump(agreement), 200


def create_new(body):
    email = body.get("email")
    user_created = False
    user = User.query.filter_by(email=email).one_or_none()
    if not user:
        user_created = True
        user = User(email=email)
        DB.session.add(user)
        DB.session.commit()

    agreement = Agreement(user_id=user.id)
    DB.session.add(agreement)
    DB.session.commit()

    response = {
        "activation_link": agreement.get_activation_link(),
        "user_created": user_created
    }
    return response, 200



from werkzeug.exceptions import abort

from app import DB
from models.user import User, UserSchema


def get_users_list():
    users = User.query.order_by(User.id).all()
    user_schema = UserSchema(many=True)
    return user_schema.dump(users)


def get_user(email):
    user = User.query.filter(User.email == email).one_or_none()
    if not user:
        abort(404, "No user exists for provided email.")

    user_schema = UserSchema()
    data = user_schema.dump(user)

    return data


def add_user(body):
    user_email = body.get("email")
    user = User.query.filter_by(email=user_email).one_or_none()
    if user:
        abort(409, "User with same email already exists.")

    user = User(email=user_email)
    DB.session.add(user)
    DB.session.commit()

    user_schema = UserSchema()
    d_user = user_schema.dump(user)

    return d_user, 201



"""
Main module of the server file
"""

import os

from flask import Response
import connexion
from flask_marshmallow import Marshmallow
from flask_sqlalchemy import SQLAlchemy

from validators import RequestBodyValidator

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

CONFIGURATION_DIRECTORY = "openapi"
DB_FILE_NAME = "sqlite3.db"

# basedir points to the directory the program is running in.
SPECIFICATION_DIR = os.path.join(BASE_DIR, CONFIGURATION_DIRECTORY)  # i.e. <current_dir>/openapi

PG_USER = "postgres"
PG_PASSWORD = "postgres"
PG_PORT = 5432
PG_DATABASE_URI = "postgresql://{username}:{password}" \
                  "@localhost:{port}/flask_testing".format(username=PG_USER,
                                                           password=PG_PASSWORD,
                                                           port=PG_PORT)
SQLITE_DATABASE_URI = 'sqlite:////' + os.path.join(BASE_DIR, DB_FILE_NAME)


def create_app(db_uri=SQLITE_DATABASE_URI):
    # Create the Connexion application instance
    _connexion_app = connexion.App(__name__, specification_dir=SPECIFICATION_DIR)

    # Get the underlying Flask app instance
    _app = _connexion_app.app  # Flask instance initialized by Connexion

    # Configure the SQLAlchemy part of the app instance
    _app.config["SQLALCHEMY_ECHO"] = True
    _app.config["SQLALCHEMY_DATABASE_URI"] = db_uri
    _app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    return _app, _connexion_app


# Get the application instance
_, connexion_app = create_app()

app = connexion_app.app  # Flask instance initialized by Connexion

DB = SQLAlchemy(app)
DB.create_all()

MA = Marshmallow(app)

# Read the specification.yml file to configure the endpoints
connexion_app.add_api("specification.yaml",
                      # strict_validation=True
                      validator_map={"body": RequestBodyValidator},
                      )


# create a URL route in our application for "/"
@connexion_app.route("/")
def home():
    return Response("Welcome to homepage.")


if __name__ == "__main__":
    connexion_app.run(debug=True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setup_requirements = list()

# with open("_setup_requirements.txt", 'r') as fh:
#     for line in fh.readlines():
#         requirement = line.strip("\n")
#         setup_requirements.append(requirement)

requirements = [
    "connexion==2.7.0"
    "connexion[openapi-ui]==0.0.6"
    "flask==1.1.1"
    "Flask-SQLAlchemy==2.4.4"
    "flask-marshmallow"
    "marshmallow"
    "marshmallow-sqlalchemy"
    "Werkzeug"
]

setuptools.setup(
    name="client_agreement_api",
    version="0.0.2",
    author="Abdullah Abid",
    author_email="abdullahabid3691@gmail.com",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Abdullahabid3691/ClientAgreement-Api",
    packages=setuptools.find_packages(include=["models", "controllers", "openapi"]),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import uuid
from datetime import datetime
from flask import request

from app import DB, MA
from models.utils import get_timestamp


class AgreementDummy:

    @staticmethod
    def generate_dummy_agreements():
        """ Populates Users + Agreements in database """

        from models.user import UserDummy
        _ite = zip(UserDummy.get_dummy_data(), [True, False, True])
        agreements = [{
            "user_id": user,
            "agreement_status": status,
            "agreed_on": get_timestamp() if status else status,
        } for user, status in list(_ite)]

        [DB.session.add(Agreement(**agreement)) for agreement in agreements]

        # commit session changes to database
        DB.session.commit()


class Agreement(DB.Model):
    __tablename__ = "agreement"  # noqa

    id = DB.Column(DB.Integer, primary_key=True)
    uuid = DB.Column('uuid', DB.Text(length=36), default=lambda: str(uuid.uuid4()))
    user_id = DB.Column(DB.Integer, DB.ForeignKey("user.id"))
    # user = DB.relationship("User", back_populates="agreement", lazy=True)
    agreement_status = DB.Column(DB.Boolean, default=False)
    agreed_on = DB.Column(DB.Integer, default=datetime.utcnow().timestamp, onupdate=datetime.utcnow().timestamp)

    def get_activation_link(self):
        host_url = request.host_url
        base_url = host_url if not host_url.endswith('/') else host_url.rstrip("/")
        api_path = request.blueprint.lstrip('/')
        activation_url = f"{base_url}/{api_path}/agreement/accept/{self.uuid}"
        return activation_url


class AgreementSchema(MA.SQLAlchemyAutoSchema):
    class Meta:
        model = Agreement
        sqla_session = DB.session  # noqa
        include_fk = True
        # include_relationships = True
        exclude = ("id",)

